@@ -1,46 +1,65 @@
 #########
-# Author:        rmp@sanger.ac.uk
-# Maintainer:    rmp@sanger.ac.uk
+# Author:        rpettett@cpan.org
+# Maintainer:    rpettett@cpan.org
 # Created:       2005-08-23
-# Last Modified: $Date: 2007/03/16 10:01:29 $ $Author: rmp $
-# Source:        $Source $
-# Id:            $Id $
+# Last Modified: $Date: 2009/01/12 10:29:26 $ $Author: rmp $
+# Source:        $Source: /cvsroot/Bio-DasLite/Bio-DasLite/lib/Bio/Das/Lite.pm,v $
+# Id:            $Id: Lite.pm,v 1.61 2009/01/12 10:29:26 rmp Exp $
 # $HeadURL $
 #
 package Bio::Das::Lite;
 use strict;
 use warnings;
-use Bio::Das::Lite::UserAgent;
+use WWW::Curl::Simple;
 use HTTP::Request;
 use HTTP::Headers;
 use SOAP::Lite;
 use Carp;
 use English qw(-no_match_vars);
+use Readonly;
 
 our $DEBUG    = 0;
-our $VERSION  = do { my @r = (q$Revision: 1.49 $ =~ /\d+/mxg); sprintf '%d.'.'%03d' x $#r, @r };
-our $BLK_SIZE = 8192;
-our $TIMEOUT  = 5;
-our $MAX_REQ  = 5;
-our $LINKRE   = qr!<link\s+href="([^"]+)"[^>]*?>([^<]*)</link>!mix;
-our $NOTERE   = qr!<note[^>]*>([^<]*)</note>!mix;
+our $VERSION  = do { my @r = (q$Revision: 2.00 $ =~ /\d+/smxg); sprintf '%d.'.'%03d' x $#r, @r };
+Readonly::Scalar our $BLK_SIZE => 8192;
+Readonly::Scalar our $TIMEOUT  => 5;
+Readonly::Scalar our $MAX_REQ  => 5;
+Readonly::Scalar our $MAX_HOST => 7;
+Readonly::Scalar our $LINKRE   => qr{<link\s+href="([^"]+)"[^>]*?>([^<]*)</link>}smix;
+Readonly::Scalar our $NOTERE   => qr{<note[^>]*>([^<]*)</note>}smix;
+
+BEGIN {
+  if(!LWP::Protocol->can('parse_head')) {
+    require LWP::Protocol;
+    *LWP::Protocol::parse_head = sub { shift->_elem('parse_head', @_); };
+  }
+}
 
 #########
 # $ATTR contains information about document structure - tags, attributes and subparts
 # This is split up by call to reduce the number of tag passes for each response
 #
 our %COMMON_STYLE_ATTRS = (
-			   'yoffset'        => [], # WTSI extension (available in Ensembl)
-			   'scorecolormin'  => [], # WTSI extension
-			   'scorecolormax'  => [], # WTSI extension
-			   'scoreheightmin' => [], # WTSI extension
-			   'scoreheightmax' => [], # WTSI extension
-			   'zindex'         => [], # WTSI extension (available in Ensembl)
-			   'height'         => [],
-			   'fgcolor'        => [],
-			   'bgcolor'        => [],
-			   'label'          => [],
-			   'bump'           => [],
+			   yoffset        => [], # WTSI extension (available in Ensembl)
+			   scorecolormin  => [], # WTSI extension
+			   scorecolormax  => [], # WTSI extension
+			   scoreheightmin => [], # WTSI extension
+			   scoreheightmax => [], # WTSI extension
+			   zindex         => [], # WTSI extension (available in Ensembl)
+			   width          => [], # WTSI extension (available in Ensembl)
+			   height         => [],
+			   fgcolor        => [],
+			   bgcolor        => [],
+			   label          => [],
+			   bump           => [],
+			  );
+our %SCORED_STYLE_ATTRS = (
+			   min            => [],
+			   max            => [],
+			   steps          => [],
+			   color1         => [],
+			   color2         => [],
+			   color3         => [],
+			   height         => [],
 			  );
 our $ATTR     = {
 		 '_segment'     => {
@@ -96,11 +115,12 @@ our $ATTR     = {
                                                                                   dbSource
                                                                                   dbVersion
                                                                                   dbAccessionId
-                                                                                  dbCoodSys)],
+                                                                                  dbCoordSys)],
 						       'alignobjectdetail' => {
 									       'alignobjectdetail' => [qw(dbSource
                                                                                                           property)],
 									      },
+						       'sequence'          => [],
 						      },
 				    'score'        => [qw(score)],
 				    'block'        => {
@@ -148,7 +168,7 @@ our $ATTR     = {
                                                                                      atomName
                                                                                      x y z)]
 								      },
-							  }, 
+							  },
 					      },
 				 'connect' => {
 					       'connect' => [qw(atomSerial type)],
@@ -160,15 +180,16 @@ our $ATTR     = {
 		 'sources' => {
 			       'source' => {
 					    'source'     => [qw(uri title doc_href description)],
-					    'maintainer' => { 
+					    'maintainer' => {
 							     'maintainer' => [qw(email)],
 							    },
 					    'version'    => {
-							     'version'     => [qw(uri created)], 
-							     'coordinates' => { 
+							     'version'     => [qw(uri created)],
+							     'coordinates' => {
 									       'coordinates' => [qw(uri
                                                                                                     source
                                                                                                     authority
+                                                                                                    taxid
                                                                                                     test_range
                                                                                                     version)],
 									      },
@@ -188,70 +209,83 @@ our $ATTR     = {
 						     'type'     => {
 								    'type'  => [qw(id)],
 								    'glyph' => {
+                                                                                'glyph'          => [qw(zoom)],
 										'arrow'          => {
-												     %COMMON_STYLE_ATTRS,
 												     'parallel'     => [],
 												     'bar_style'    => [], # WTSI extension
+												     %COMMON_STYLE_ATTRS,
 												    },
 										'anchored_arrow' => {
-												     %COMMON_STYLE_ATTRS,
 												     'parallel'     => [],
 												     'orientation'  => [], # WTSI extension
 												     'no_anchor'    => [], # WTSI extension
 												     'bar_style'    => [], # WTSI extension
+												     %COMMON_STYLE_ATTRS,
 												    },
 										'box'            => {
-												     %COMMON_STYLE_ATTRS,
 												     'linewidth'    => [],
 												     'pattern'      => [],  # WTSI extension
+												     %COMMON_STYLE_ATTRS,
 												    },
 										'farrow'         => {                      # WTSI extension
-												     %COMMON_STYLE_ATTRS,
 												     'orientation'  => [],
 												     'no_anchor'    => [],
 												     'bar_style'    => [], # WTSI extension
+												     %COMMON_STYLE_ATTRS,
 												    },
 										'rarrow'         => {                      # WTSI extension
-												     %COMMON_STYLE_ATTRS,
 												     'orientation'  => [],
 												     'no_anchor'    => [],
 												     'bar_style'    => [], # WTSI extension
+												     %COMMON_STYLE_ATTRS,
 												    },
 										'cross'          => {
-												     %COMMON_STYLE_ATTRS,
 												     'linewidth'    => [],  # WTSI extension
+												     %COMMON_STYLE_ATTRS,
 												    },
 										'dot'            => \%COMMON_STYLE_ATTRS,
 										'ex'             => {
-												     %COMMON_STYLE_ATTRS,
 												     'linewidth'    => [],  # WTSI extension
+												     %COMMON_STYLE_ATTRS,
 												    },
 										'hidden'         => \%COMMON_STYLE_ATTRS,
 										'line'           => {
-												     %COMMON_STYLE_ATTRS,
 												     'style'        => [],
+												     %COMMON_STYLE_ATTRS,
 												    },
 										'span'           => {
-												     %COMMON_STYLE_ATTRS,
 												     'bar_style'    => [], # WTSI extension
+												     %COMMON_STYLE_ATTRS,
 												    },
 										'text'           => {
-												     %COMMON_STYLE_ATTRS,
 												     'font'         => [],
 												     'fontsize'     => [],
 												     'string'       => [],
 												     'style'        => [],
+												     %COMMON_STYLE_ATTRS,
 												    },
 										'primers'        => \%COMMON_STYLE_ATTRS,
 										'toomany'        => {
-												     %COMMON_STYLE_ATTRS,
 												     'linewidth'    => [],
+												     %COMMON_STYLE_ATTRS,
 												    },
 										'triangle'       => {
-												     %COMMON_STYLE_ATTRS,
 												     'linewidth'    => [],
 												     'direction'    => [],
 												     'orientation'  => [],
+												     %COMMON_STYLE_ATTRS,
+												    },
+										'gradient'       => {
+												     %SCORED_STYLE_ATTRS,
+												    },
+										'histogram'      => {
+												     %SCORED_STYLE_ATTRS,
+												    },
+										'tiling'         => {
+												     %SCORED_STYLE_ATTRS,
+												    },
+										'lineplot'       => {
+												     %SCORED_STYLE_ATTRS,
 												    },
 									       },
 								   },
@@ -290,7 +324,7 @@ sub new {
 
   if($ref && ref $ref) {
     for my $arg (qw(dsn timeout caching callback registry user_agent
-                    http_proxy proxy_user proxy_pass)) {
+                    http_proxy proxy_user proxy_pass no_proxy)) {
       if(exists $ref->{$arg} && $self->can($arg)) {
 	$self->$arg($ref->{$arg});
       }
@@ -310,6 +344,7 @@ sub new_from_registry {
   return $self;
 }
 
+# We implement this method because LWP does not parse user/password
 sub http_proxy {
   my ($self, $proxy) = @_;
   if($proxy) {
@@ -321,7 +356,7 @@ sub http_proxy {
     $self->{'_checked_http_proxy_env'} = 1;
   }
 
-  if($self->{'http_proxy'} =~ m|^(https?://)(\S+):(.*?)\@(.*?)$|mx) {
+  if($self->{'http_proxy'} =~ m{^(https?://)(\S+):(.*?)\@(.*?)$}smx) {
     #########
     # http_proxy contains username & password - we'll set them up here:
     #
@@ -334,11 +369,29 @@ sub http_proxy {
   return $self->{'http_proxy'};
 }
 
+sub no_proxy {
+  my ($self, @args) = @_;
+
+  if (scalar @args) {
+    if ($args[0] && ref $args[0] && ref $args[0] eq 'ARRAY') {
+      $self->{'no_proxy'} = $args[0];
+    } else {
+      $self->{'no_proxy'} = \@args;
+    }
+  }
+
+  if(!$self->{'_checked_no_proxy_env'}) {
+    $self->{'no_proxy'} ||= [split /\s*,\s*/smx, $ENV{'no_proxy'} || q()];
+    $self->{'_checked_no_proxy_env'} = 1;
+  }
+
+  return $self->{'no_proxy'} || [];
+}
+
 sub _get_set {
   my ($self, $key, $value) = @_;
   if(defined $value) {
     $self->{$key} = $value;
-#print {*STDERR} qq(set self $key = @{[$value||'undef']}\n);
   }
   return $self->{$key};
 }
@@ -368,6 +421,11 @@ sub caching {
   return $self->_get_set('caching', $val);
 }
 
+sub max_hosts {
+  my ($self, $val) = @_;
+  return $self->_get_set('_max_hosts', $val);
+}
+
 sub max_req {
   my ($self, $val) = @_;
   return $self->_get_set('_max_req', $val);
@@ -385,7 +443,7 @@ sub basename {
   my @res          = ();
 
   for my $service (@dsns) {
-    $service =~ m|(https?://.*/das)/?|mx;
+    $service =~ m{(https?://.*/das)/?}smx;
     if($1) {
       push @res, $1;
     }
@@ -486,10 +544,16 @@ sub _generic_request {
   delete $self->{'currentsegs'};
   my $results   = {};
   my $reqname   = $fname;
-  $reqname      =~ s/[\(\)]//mxg;
-  ($fname)      = $fname =~ /^([a-z_]+)/mx;
+  $reqname      =~ s/[\(\)]//smxg;
+  ($fname)      = $fname =~ /^([a-z_]+)/smx;
 
-  my $ref       = $self->build_requests($query, $fname, $reqname, $opts, $results);
+  my $ref       = $self->build_requests({
+					 query   => $query,
+					 fname   => $fname,
+					 reqname => $reqname,
+					 opts    => $opts,
+					 results => $results
+					});
 
   $self->_fetch($ref, $opts->{'headers'});
   $DEBUG and print {*STDERR} qq(Content retrieved\n);
@@ -503,7 +567,7 @@ sub _generic_request {
     $DEBUG and print {*STDERR} qq(Performing cache handling\n);
     for my $s (keys %{$results}) {
       if($DEBUG && !$results->{$s}) {
-	print {*STDERR} qq(CACHE HIT for $s\n);
+	print {*STDERR} qq(CACHE HIT for $s\n); ## no critic (InputOutput::RequireCheckedSyscalls)
       }
       $results->{$s}          ||= $self->{'_cache'}->{$s};
       $self->{'_cache'}->{$s} ||= $results->{$s};
@@ -541,7 +605,7 @@ sub build_queries {
 	#########
 	# ... or if the first array arg is a hash, stitch the series of queries together
 	#
-	push @queries, map {
+	push @queries, map { ## no critic (ProhibitComplexMappings)
 	  my $q = $_;
 	  join q(;), map { "$_=$q->{$_}" } grep { $q->{$_} } @{$OPTS->{$fname}};
 	} @{$query};
@@ -583,13 +647,12 @@ sub _hack_fname {
 }
 
 sub build_requests {
-  my ($self,
-      $query,
-      $fname,
-      $reqname,
-      $opts,
-      $results) = @_;
-
+  my ($self, $args) = @_;
+  my $query     = $args->{query};
+  my $fname     = $args->{fname};
+  my $reqname   = $args->{reqname};
+  my $opts      = $args->{opts};
+  my $results   = $args->{results};
   my $queries   = $self->build_queries($query, $fname);
   my $attr      = $ATTR->{$fname};
   my $dsn       = $opts->{'use_basename'}?$self->basename():$self->dsn();
@@ -623,22 +686,23 @@ sub build_requests {
 	  # If we haven't yet found segment information for this request
 	  # Then look for some. This one is a non-destructive scan.
 	  #
-	  my $matches = $self->{'data'}->{$request}  =~ m!(<segment[^>]*>)!mix;
+	  my $matches = $self->{'data'}->{$request}  =~ m{(<segment[^>]*>)}smix;
 
 	  if($matches) {
 	    my $seginfo = [];
-	    _parse_branch($self,
-			  $request,
-			  $seginfo,
-			  $ATTR->{'_segment'},
-			  $1,
-			  0);
+	    $self->_parse_branch({
+				  request    => $request,
+				  seginfo    => $seginfo,
+				  attr       => $ATTR->{'_segment'},
+				  blk        => $1,
+				  addseginfo => 0,
+				 });
 	    $self->{'currentsegs'}->{$request} = $seginfo->[0];
 	  }
 	}
 
 	if($DEBUG) {
-	  print {*STDERR} qq(invoking _parse_branch for $fname\n);
+	  print {*STDERR} qq(invoking _parse_branch for $fname\n); ## no critic (InputOutput::RequireCheckedSyscalls)
 	}
 
 	#########
@@ -649,13 +713,19 @@ sub build_requests {
 	}
 	$fname = $self->_hack_fname($fname);
 
-	my $pat = qr!(<$fname.*?/$fname>|<$fname[^>]+/>)!smix;
-	while($self->{'data'}->{$request} =~ s/$pat//mx) {
-	  _parse_branch($self, $request, $results->{$request}, $attr, $1, 1);
+	my $pat = qr{(<$fname.*?/$fname>|<$fname[^>]+/>)}smix;
+	while($self->{'data'}->{$request} =~ s/$pat//smx) {
+	  $self->_parse_branch({
+				request    => $request,
+				seginfo    => $results->{$request},
+				attr       => $attr,
+				blk        => $1,
+				addseginfo => 1,
+			       });
 	}
 
 	if($DEBUG) {
-	  print {*STDERR} qq(completed _parse_branch\n);
+	  print {*STDERR} qq(completed _parse_branch\n); ## no critic (InputOutput::RequireCheckedSyscalls)
 	}
 
 	return;
@@ -716,18 +786,21 @@ sub postprocess {
 
 #########
 # Set up the parallel HTTP fetching
-# This uses our LWP::Parallel::UserAgent subclass which has better proxy handling
+# This uses our LWP::Parallel::UserAgent subclass which handles DAS statuses
 #
 sub _fetch {
   my ($self, $url_ref, $headers) = @_;
-  $self->{'ua'}                ||= Bio::Das::Lite::UserAgent->new(
-								'http_proxy' => $self->http_proxy(),
-							       );
-  $self->{'ua'}->initialize();
-  $self->{'ua'}->max_req($self->max_req()||$MAX_REQ);
-  $self->{'statuscodes'}          = {};
-  $headers                      ||= {};
-  $headers->{'X-Forwarded-For'} ||= $ENV{'HTTP_X_FORWARDED_FOR'};
+
+  my $ua = WWW::Curl::Simple->new;
+
+  $self->{'statuscodes'} = {};
+  if(!$headers) {
+    $headers = {};
+  }
+
+  if($ENV{HTTP_X_FORWARDED_FOR}) {
+    $headers->{'X-Forwarded-For'} ||= $ENV{'HTTP_X_FORWARDED_FOR'};
+  }
 
   for my $url (keys %{$url_ref}) {
     if(ref $url_ref->{$url} ne 'CODE') {
@@ -735,37 +808,25 @@ sub _fetch {
     }
     $DEBUG and print {*STDERR} qq(Building HTTP::Request for $url [timeout=$self->{'timeout'}] via $url_ref->{$url}\n);
 
-    my $headers = HTTP::Headers->new(%{$headers});
-    $headers->user_agent($self->user_agent());
+    my $http_headers = HTTP::Headers->new(%{$headers});
+    $http_headers->user_agent($self->user_agent());
 
     if($self->proxy_user() && $self->proxy_pass()) {
       $headers->proxy_authorization_basic($self->proxy_user(), $self->proxy_pass());
     }
 
-    my $response = $self->{'ua'}->register(HTTP::Request->new('GET', $url, $headers),
-					   $url_ref->{$url},
-					   $BLK_SIZE);
-    if($response) {
-      $self->{'statuscodes'}->{$url} ||= $response->status_line();
-    }
+    $ua->register(HTTP::Request->new('GET', $url, $http_headers));
   }
 
+  my @res = $ua->perform;
   $DEBUG and print {*STDERR} qq(Requests submitted. Waiting for content\n);
-  eval {
-    $self->{'ua'}->wait($self->{'timeout'});
-  };
-
-  if($EVAL_ERROR) {
-    carp $EVAL_ERROR;
+  for my $req (@res) {
+    my $res              = $req->response;
+    my $uri = $res->request->uri;
+    $self->{statuscodes}->{$uri} = $res->code;
+    $url_ref->{$uri}->($res->content);
   }
 
-  for my $url (keys %{$url_ref}) {
-    if(ref $url_ref->{$url} ne 'CODE') {
-      next;
-    }
-
-    $self->{'statuscodes'}->{$url} ||= '200';
-  }
   return;
 }
 
@@ -773,14 +834,14 @@ sub statuscodes {
   my ($self, $url)         = @_;
   $self->{'statuscodes'} ||= {};
 
-  if($self->{'ua'}) {
-    my $uacodes = $self->{'ua'}->statuscodes();
-    for my $k (keys %{$uacodes}) {
-      if($uacodes->{$k}) {
-	$self->{'statuscodes'}->{$k} = $uacodes->{$k};
-      }
-    }
-  }
+#  if($self->{'ua'}) {
+#    my $uacodes = $self->{'ua'}->statuscodes();
+#    for my $k (keys %{$uacodes}) {
+#      if($uacodes->{$k}) {
+#	$self->{'statuscodes'}->{$k} = $uacodes->{$k};
+#      }
+#    }
+#  }
 
   return $url?$self->{'statuscodes'}->{$url}:$self->{'statuscodes'};
 }
@@ -790,9 +851,14 @@ sub statuscodes {
 # recursively parse the XML blocks and build the corresponding response data structure
 #
 sub _parse_branch {
-  my ($self, $dsn, $ar_ref, $attr, $blk, $addseginfo, $depth) = @_;
-  $depth ||= 0;
-  my $ref  = {};
+  my ($self, $args) = @_;
+  my $dsn           = $args->{request};
+  my $ar_ref        = $args->{seginfo};
+  my $attr          = $args->{attr};
+  my $blk           = $args->{blk};
+  my $addseginfo    = $args->{addseginfo};
+  my $depth         = $args->{depth} || 0;
+  my $ref           = {};
 
   my (@parts, @subparts);
   while(my ($k, $v) = each %{$attr}) {
@@ -809,9 +875,16 @@ sub _parse_branch {
   for my $subpart (@subparts) {
     my $subpart_ref  = [];
 
-    my $pat = qr!(<$subpart[^>]*/>|<$subpart[^>]*?(?\!/)>.*?/$subpart>)!smix;
-    while($blk =~ s/$pat//mx) {
-      _parse_branch($self, $dsn, $subpart_ref, $attr->{$subpart}, $1, 0, $depth+1);
+    my $pat = qr{(<$subpart[^>]*/>|<$subpart[^>]*?(?!/)>.*?/$subpart>)}smix;
+    while($blk =~ s/$pat//smx) {
+      $self->_parse_branch({
+			    request    => $dsn,
+			    seginfo    => $subpart_ref,
+			    attr       => $attr->{$subpart},
+			    blk        => $1,
+			    addseginfo => 0,
+			    depth      => $depth+1,
+			   });
     }
 
     if(scalar @{$subpart_ref}) {
@@ -832,13 +905,13 @@ sub _parse_branch {
     my $opts = $attr->{$tag}||[];
 
     for my $a (@{$opts}) {
-      ($tmp)              = $blk =~ m|<$tag[^>]+$a="([^"]+?)"|smix;
+      ($tmp)              = $blk =~ m{<$tag[^>]+$a="([^"]+?)"}smix;
       if(defined $tmp) {
 	$ref->{"${tag}_$a"} = $tmp;
       }
     }
 
-    ($tmp) = $blk =~ m|<$tag[^>]*>([^<]+)</$tag>|smix;
+    ($tmp) = $blk =~ m{<$tag[^>]*>([^<]+)</$tag>}smix;
     if(defined $tmp) {
       $tmp         =~ s/^\s+$//smgx;
       if(length $tmp) {
@@ -846,7 +919,7 @@ sub _parse_branch {
       }
     }
     if($tmp && $DEBUG) {
-      print {*STDERR} q( )x($depth*2), qq(  $tag = $tmp\n);
+      print {*STDERR} q( )x($depth*2), qq(  $tag = $tmp\n); ## no critic (InputOutput::RequireCheckedSyscalls)
     }
   }
 
@@ -874,19 +947,19 @@ sub _parse_twig {
   #########
   # handle multiples of twig elements here
   #
-  $blk =~ s!$LINKRE!{
+  $blk =~ s/$LINKRE/{
                      $ref->{'link'} ||= [];
                      push @{$ref->{'link'}}, {
                                               'href' => $1,
                                               'txt'  => $2,
                                              };
                      q()
-                    }!smegix;
-  $blk =~ s!$NOTERE!{
+                    }/smegix;
+  $blk =~ s/$NOTERE/{
                      $ref->{'note'} ||= [];
                      push @{$ref->{'note'}}, $1;
                      q()
-                    }!smegix;
+                    }/smegix;
 
   if($addseginfo && $self->{'currentsegs'}->{$dsn}) {
     while(my ($k, $v) = each %{$self->{'currentsegs'}->{$dsn}}) {
@@ -938,11 +1011,16 @@ sub registry_sources {
 
       $DEBUG and print {*STDERR} qq(Running request for $reg\n);
 
-      $SIG{ALRM} = sub { croak 'timeout'; };
-      alarm $TIMEOUT;
+      local $SIG{ALRM} = sub { croak 'timeout'; };
+      alarm $self->timeout();
+
       eval {
 	push @{$self->{'_registry_sources'}}, @{$soap->listServices()};
+
+      } or do {
+	carp $EVAL_ERROR;
       };
+
       alarm 0;
     }
   }
@@ -962,7 +1040,7 @@ sub registry_sources {
   if((ref $capability eq 'ARRAY') &&
      (scalar @{$capability})) {
     my $str    = join q(|), @{$capability};
-    my $match  = qr/$str/;
+    my $match  = qr/$str/smx;
     $sources = [grep { $self->_filter_capability($_, $match) } @{$sources}];
   }
 
@@ -991,7 +1069,7 @@ sub _filter_category {
   my ($self, $src, $match) = @_;
   for my $scoord (@{$src->{'coordinateSystem'}}) {
     for my $m (@{$match}) {
-      return 1 if($scoord->{'category'} eq $m);
+       return 1 if($scoord->{'category'} eq $m);
     }
   }
   return 0;
@@ -1030,6 +1108,7 @@ Bio::Das::Lite - Perl extension for the DAS (HTTP+XML) Protocol (http://biodas.o
  Options can be: dsn        (optional scalar or array ref, URLs of DAS services)
                  timeout    (optional int,      HTTP fetch timeout in seconds)
                  http_proxy (optional scalar,   web cache or proxy if not set in %ENV)
+                 no_proxy   (optional list/ref, non-proxiable domains if not set in %ENV)
                  caching    (optional bool,     primitive caching on/off)
                  callback   (optional code ref, callback for processed XML blocks)
                  registry   (optional array ref containing DAS registry service URLs
@@ -1074,6 +1153,14 @@ For a complete list of capabilities and categories, see:
 
     $das->proxy_pass('secretpassword');
 
+=head2 no_proxy : Get/Set domains to not use proxy for
+
+    $das->no_proxy('ebi.ac.uk', 'localhost');
+    OR
+    $das->no_proxy( ['ebi.ac.uk', 'localhost'] );
+    
+    Always returns an arrayref
+
 =head2 user_agent : Get/Set user-agent for request headers
 
     $das->user_agent('GroovyDAS/1.0');
@@ -1164,8 +1251,9 @@ For a complete list of capabilities and categories, see:
                                       # optional args - see DAS Spec
                                      });
   my $feature_data4 = $das->features([
-                                      {'segment' => '1:1,1000000','type' => 'karyotype',},
-                                      {'segment' => '2:1,1000000',},
+                                      {'segment'  => '1:1,1000000','type' => 'karyotype',},
+                                      {'segment'  => '2:1,1000000',},
+                                      {'group_id' => 'OTTHUMG00000036084',},
                                      ]);
 
   #########
@@ -1185,6 +1273,9 @@ For a complete list of capabilities and categories, see:
   # or:
   $das->features(['1:1,1000000', '2:1,1000000', '3:1,1000000'], $callback);
 
+  # or:
+  $das->features([{'group_id' => 'OTTHUMG00000036084'}, '2:1,1000000', '3:1,1000000'], $callback);
+
 =head2 alignment : Retrieve protein alignment data for a query.  This can be a multiple sequence alignment
                     or pairwise alignment.  Note - this has not been tested for structural alignments as there
                     is currently no Das source avialable.
@@ -1213,7 +1304,12 @@ For a complete list of capabilities and categories, see:
   my $code         = $das->statuscodes($url);
   my $code_hashref = $das->statuscodes();
 
-=head2 max_req set number of running concurrent requests
+=head2 max_hosts set number of running concurrent host connections
+
+  $das->max_hosts(7);
+  print $das->max_hosts();
+
+=head2 max_req set number of running concurrent requests per host
 
   $das->max_req(5);
   print $das->max_req();
@@ -1270,6 +1366,8 @@ This module is an implementation of a client for the DAS protocol (XML over HTTP
 
 =head1 INCOMPATIBILITIES
 
+Versions of LWP from 5.816 onwards (at time of writing upto 5.820) break LWP::Parallel and
+therefore Bio::Das::Lite. It is recommended to use version 5.814 of LWP.
 
 =head1 BUGS AND LIMITATIONS
 
@@ -1288,7 +1386,7 @@ The DAS Registry at:
 
 =head1 AUTHOR
 
-Roger Pettett, E<lt>rmp@sanger.ac.ukE<gt>
+Roger Pettett, E<lt>rpettett@cpan.orgE<gt>
 
 =head1 LICENSE AND COPYRIGHT
 
